/*
 * hid.c  -- HID (Human Interface Device)
 *
 * $Id: hid.c,v 1.2 2004/02/08 02:03:52 hos Exp $
 *
 */


#include "main.h"


#define REPORT_DESCRIPTOR_SIZE (sizeof(hid_report_descriptor))

static BYTE hid_report_descriptor[] = {
    /* joystick */
    0x05, 0x01,                 /* Usage Page(Generic Desktop) */
    0x09, 0x04,                 /* Usage(Joystick) */
    0xa1, 0x01,                 /* Collection(Application) */

    /* x, y */
    0x05, 0x01,                 /*   Usage Page(Generic Desktop) */
    0x09, 0x01,                 /*   Usage(Pointer) */
    0xa1, 0x00,                 /*   Collection(Physical) */
    0x15, 0x00,                 /*     Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*     Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*     Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*     Physical Maximum(0x00ff) */
    0x75, 0x08,                 /*     Report Size(8) */
    0x95, 0x02,                 /*     Report Count(2) */
    0x05, 0x01,                 /*     Usage Page(Generic Desktop) */
    0x09, 0x30,                 /*     Usage(X) */
    0x09, 0x31,                 /*     Usage(Y) */
    0x81, 0x02,                 /*     Input(Variable, Absolute) */
    0xc0,                       /*   End Collection */

    /* rx, ry */
    0x05, 0x01,                 /*   Usage Page(Generic Desktop) */
    0x09, 0x01,                 /*   Usage(Pointer) */
    0xa1, 0x00,                 /*   Collection(Physical) */
    0x15, 0x00,                 /*     Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*     Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*     Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*     Physical Maximum(0x00ff) */
    0x75, 0x08,                 /*     Report Size(8) */
    0x95, 0x02,                 /*     Report Count(2) */
    0x05, 0x01,                 /*     Usage Page(Generic Desktop) */
    0x09, 0x33,                 /*     Usage(Rx) */
    0x09, 0x34,                 /*     Usage(Ry) */
    0x81, 0x02,                 /*     Input(Variable, Absolute) */
    0xc0,                       /*   End Collection */

    /* rz */
    0x05, 0x01,                 /*   Usage Page(Generic Desktop) */
    0x09, 0x01,                 /*   Usage(Pointer) */
    0xa1, 0x00,                 /*   Collection(Physical) */
    0x15, 0x00,                 /*     Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*     Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*     Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*     Physical Maximum(0x00ff) */
    0x75, 0x08,                 /*     Report Size(8) */
    0x95, 0x01,                 /*     Report Count(1) */
    0x05, 0x01,                 /*     Usage Page(Generic Desktop) */
    0x09, 0x35,                 /*     Usage(Rz) */
    0x81, 0x02,                 /*     Input(Variable, Absolute) */
    0xc0,                       /*   End Collection */

    /* throttle */
    0x75, 0x08,                 /*   Report Size(8) */
    0x95, 0x01,                 /*   Report Count(1) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*   Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*   Physical Maximum(0x00ff) */
    0x05, 0x02,                 /*   Usage Page(Simulation) */
    0x09, 0xbb,                 /*   Usage(Throttle) */
    0x81, 0x02,                 /*   Input(Variable, Absolute) */

    /* hat switch 1 */
    0x75, 0x04,                 /*   Report Size(4) */
    0x95, 0x01,                 /*   Report Count(1) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x25, 0x07,                 /*   Logical Maximum(0x07) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x46, 0x3b, 0x01,           /*   Physical Maximum(0x013b) */
    0x65, 0x14,                 /*   Unit(Degree) */
    0x05, 0x01,                 /*   Usage Page(Generic Desktop) */
    0x09, 0x39,                 /*   Usage(Hat Switch) */
    0x81, 0x42,                 /*   Input(Variable, Absolute, Null State) */

    /* hat switch 2 */
    0x75, 0x04,                 /*   Report Size(4) */
    0x95, 0x01,                 /*   Report Count(1) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x25, 0x07,                 /*   Logical Maximum(0x07) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x46, 0x3b, 0x01,           /*   Physical Maximum(0x013b) */
    0x65, 0x14,                 /*   Unit(Degree) */
    0x05, 0x01,                 /*   Usage Page(Generic Desktop) */
    0x09, 0x39,                 /*   Usage(Hat Switch) */
    0x81, 0x42,                 /*   Input(Variable, Absolute, Null State) */

    /* digital button x 16 */
    0x75, 0x01,                 /*   Report Size(1) */
    0x95, 0x10,                 /*   Report Count(16) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x25, 0x01,                 /*   Logical Maximum(0x01) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x45, 0x01,                 /*   Physical Maximum(0x01) */
    0x05, 0x09,                 /*   Usage Page(Button) */
    0x19, 0x01,                 /*   Usage Minimum(Button 1) */
    0x29, 0x10,                 /*   Usage Maximum(Button 16) */
    0x81, 0x02,                 /*   Input(Variable, Absolute) */

    /* analog button x 2 */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x26, 0xff, 0x00,           /*   Logical Maximum(0x00ff) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x46, 0xff, 0x00,           /*   Physical Maximum(0x00ff) */
    0x05, 0x09,                 /*   Usage Page(Button) */

    0x75, 0x08,                 /*   Report Size(8) */
    0x95, 0x01,                 /*   Report Count(1) */
    0x09, 0x11,                 /*   Usage(Button 17) */
    0x81, 0x02,                 /*   Input(Variable, Absolute) */

    0x75, 0x08,                 /*   Report Size(8) */
    0x95, 0x01,                 /*   Report Count(1) */
    0x09, 0x12,                 /*   Usage(Button 18) */
    0x81, 0x02,                 /*   Input(Variable, Absolute) */

    /* Output Report */
    0x75, 0x08,                 /*   Report Size(8) */
    0x95, 0x25,                 /*   Report Count(1) */
    0x15, 0x00,                 /*   Logical Minimum(0x00) */
    0x25, 0x0f,                 /*   Logical Maximum(0xff) */
    0x35, 0x00,                 /*   Physical Minimum(0x00) */
    0x45, 0x0f,                 /*   Physical Maximum(0xff) */
    0x06, 0x00, 0xff,           /*   Usage Page(Vendor defined) */
    0x09, 0x01,                 /*   Usage(Vendor defined 1) */
    0x91, 0x02,                 /*   Output(Variable, Absolute) */

    0xc0                        /* End Collection */
};


NTSTATUS hid_init(IN PDEVICE_OBJECT dev_obj)
{
    device_extension_t *dev_ext;

    DBG_ENTER(("%p", dev_obj));

    dev_ext = GET_DEV_EXT(dev_obj);

    dev_ext->report_size = REPORT_DESCRIPTOR_SIZE;
    RtlZeroMemory(&dev_ext->input_state, sizeof(dev_ext->input_state));
    RtlZeroMemory(&dev_ext->output_state, sizeof(dev_ext->output_state));
    KeInitializeSpinLock(&dev_ext->state_lock);

    DBG_LEAVE(("STATUS_SUCCESS"));
    return STATUS_SUCCESS;
}


NTSTATUS hid_get_rep_desc(IN PDEVICE_OBJECT dev_obj,
                          OUT BYTE *buf, OUT ULONG *size)
{
    DBG_ENTER(("%p, %p, %u", dev_obj, buf, *size));

    if(*size < REPORT_DESCRIPTOR_SIZE) {
        DBG_LEAVE(("STATUS_BUFFER_TOO_SMALL"));
        return STATUS_BUFFER_TOO_SMALL;
    }

    RtlCopyMemory(buf, hid_report_descriptor, REPORT_DESCRIPTOR_SIZE);
    *size = REPORT_DESCRIPTOR_SIZE;

    DBG_LEAVE(("STATUS_SUCCESS"));
    return STATUS_SUCCESS;
}


NTSTATUS hid_read_rep(IN PDEVICE_OBJECT dev_obj,
                      OUT BYTE *buf, OUT ULONG *size)
{
    struct controller_input_report in_state;

    DBG_ENTER(("%p, %p, %u", dev_obj, buf, *size));

    if(*size < sizeof(in_state)) {
        DBG_LEAVE(("STATUS_BUFFER_TOO_SMALL"));
        return STATUS_BUFFER_TOO_SMALL;
    }

    in_state = GET_DEV_EXT(dev_obj)->input_state;

    RtlCopyMemory(buf, &in_state, sizeof(in_state));
    *size = sizeof(in_state);

    DBG_LEAVE(("STATUS_SUCCESS"));
    return STATUS_SUCCESS;
}


NTSTATUS hid_write_rep(IN PDEVICE_OBJECT dev_obj,
                       IN BYTE *buf, IN ULONG *size)
{
    device_extension_t *dev_ext;
    struct controller_output_report out_state;
    NTSTATUS nt_ret;

    DBG_ENTER(("%p, %p, %u", dev_obj, buf, *size));

    dev_ext = GET_DEV_EXT(dev_obj);

    if(*size > sizeof(out_state)) {
        *size = sizeof(out_state);
    }

    RtlZeroMemory(&out_state, sizeof(out_state));
    RtlCopyMemory(&out_state, buf, *size);

    KeAcquireSpinLockAtDpcLevel(&dev_ext->state_lock); /* Lock */
    dev_ext->output_state = out_state;
    KeReleaseSpinLockFromDpcLevel(&dev_ext->state_lock); /* Unlock */

    {
        struct controller_raw_output raw_out;
        ULONG raw_size;

        hid_generate_raw_output(&out_state, &raw_out);

        DBG_OUT(("write report: %02x", raw_out.vr_0c.viblation));
        raw_size = sizeof(raw_out.vr_0c);
        nt_ret = usb_put_vendor_request(dev_obj, &raw_out.vr_0c, &raw_size,
                                        0x0c, 0x01);
    }

    DBG_LEAVE(("%d", nt_ret));
    return nt_ret;
}


VOID hid_update_from_raw_input(IN PDEVICE_OBJECT dev_obj,
                               IN BYTE *buf, IN ULONG size)
{
    device_extension_t *dev_ext;
    struct controller_raw_input *org_raw_in, raw_in;
    struct controller_input_report in_state;

    const static UCHAR hat_map[] = {
       /* 0    0    1    1 */
       /* 0    1    0    1 */
        0xf, 0xf, 0xf, 0x5, /* 00 */
        0xf, 0xf, 0x7, 0x6, /* 01 */
        0xf, 0x3, 0xf, 0x4, /* 10 */
        0x1, 0x2, 0x0, 0xf  /* 11 */
    };
    const static UCHAR mode_map[] = {
        0x00,
        0x02,                   /* M2 */
        0x01,                   /* M3 */
        0x04                    /* M1 */
    };

    DBG_ENTER(("%p, %p, %u", dev_obj, buf, size));

    dev_ext = GET_DEV_EXT(dev_obj);

    if(size < sizeof(struct controller_raw_input)) {
        DBG_LEAVE(("void: size too small"));
        return;
    }

    org_raw_in = (struct controller_raw_input *)buf;
    raw_in = *org_raw_in;

    RtlZeroMemory(&in_state, sizeof(in_state));
    in_state.x = raw_in.ir.stick_x;
    in_state.y = raw_in.ir.stick_y;
    in_state.rx = raw_in.ir.hat_x;
    in_state.ry = raw_in.ir.hat_y;
    in_state.rz = raw_in.ir.rudder;
    in_state.throttle = 0xff - raw_in.ir.throttle;

    in_state.hat1 = hat_map[((raw_in.vr_00.dpad1_left << 3) |
                             (raw_in.vr_00.dpad1_bottom << 2) |
                             (raw_in.vr_00.dpad1_right << 1) |
                             (raw_in.vr_00.dpad1_top << 0))];
    in_state.hat2 = hat_map[((raw_in.vr_01.dpad2_left << 3) |
                             (raw_in.vr_01.dpad2_bottom << 2) |
                             (raw_in.vr_01.dpad2_right << 1) |
                             (raw_in.vr_01.dpad2_top << 0))];

    in_state.digital_button[0] = ~((raw_in.vr_00.trigger << 0) |
                                   (raw_in.vr_00.fire_c << 1) |
                                   (raw_in.vr_00.launch << 2) |
                                   (raw_in.vr_00.button_d << 3) |
                                   (raw_in.vr_00.button_st << 4) |
                                   (raw_in.vr_01.button_sw1 << 5) |
                                   (raw_in.vr_00.hat << 6) |
                                   (raw_in.vr_01.dpad3_left << 7));
    in_state.digital_button[1] = (((! raw_in.vr_01.dpad3_middle) << 0) |
                                  ((! raw_in.vr_01.dpad3_right) << 1) |
                                  (mode_map[raw_in.vr_01.mode_select] << 2));

    in_state.analog_button[0] = 0xff - raw_in.ir.button_a;
    in_state.analog_button[1] = 0xff - raw_in.ir.button_b;

    dev_ext->input_state = in_state;

    DBG_LEAVE(("void"));
}

VOID hid_generate_raw_output(struct controller_output_report *out_state,
                             struct controller_raw_output *raw_out)
{
    DBG_ENTER(("%p, %p", out_state, raw_out));

    RtlZeroMemory(raw_out, sizeof(*raw_out));

    raw_out->vr_0c.viblation = out_state->viblation;

    DBG_LEAVE(("void"));
}
