/*
 * main.c  -- entry point and main procedures
 *
 * $Id: main.c,v 1.1 2004/02/08 01:15:21 hos Exp $
 *
 */


#include "main.h"


NTSTATUS DDKAPI device_add(IN PDRIVER_OBJECT drv_obj,
                           IN PDEVICE_OBJECT phys_dev_obj);
VOID DDKAPI driver_unload(IN PDRIVER_OBJECT drv_obj);
NTSTATUS DDKAPI irp_create(IN PDEVICE_OBJECT dev_obj, IN PIRP irp);
NTSTATUS DDKAPI irp_close(IN PDEVICE_OBJECT dev_obj, IN PIRP irp);
NTSTATUS DDKAPI irp_pnp(IN PDEVICE_OBJECT dev_obj, IN PIRP irp);
NTSTATUS DDKAPI irp_power(IN PDEVICE_OBJECT dev_obj, IN PIRP irp);
NTSTATUS DDKAPI irp_inter_ioctl(IN PDEVICE_OBJECT dev_obj, IN PIRP irp);

NTSTATUS inc_req_cnt(IN device_extension_t *dev_ext);
VOID dec_req_cnt(IN device_extension_t *dev_ext);

VOID load_settings(IN PDEVICE_OBJECT dev_obj);


/* entry point of device driver */
NTSTATUS DDKAPI DriverEntry(IN PDRIVER_OBJECT drv_obj,
                            IN PUNICODE_STRING reg_path)
{
    HID_MINIDRIVER_REGISTRATION hid_mdrv_regist;
    NTSTATUS nt_ret;

    DBG_ENTER(("%p, \"%ws\"", drv_obj, reg_path->Buffer));

    nt_ret = STATUS_SUCCESS;

    drv_obj->DriverExtension->AddDevice = device_add;
    drv_obj->DriverUnload = driver_unload;
    drv_obj->MajorFunction[IRP_MJ_CREATE] = irp_create;
    drv_obj->MajorFunction[IRP_MJ_CLOSE] = irp_close;
    drv_obj->MajorFunction[IRP_MJ_PNP] = irp_pnp;
    drv_obj->MajorFunction[IRP_MJ_POWER] = irp_power;
    drv_obj->MajorFunction[IRP_MJ_INTERNAL_DEVICE_CONTROL] = irp_inter_ioctl;

    RtlZeroMemory(&hid_mdrv_regist, sizeof(hid_mdrv_regist));
    hid_mdrv_regist.Revision = HID_REVISION;
    hid_mdrv_regist.DriverObject = drv_obj;
    hid_mdrv_regist.RegistryPath = reg_path;
    hid_mdrv_regist.DevicesArePolled = TRUE;
    hid_mdrv_regist.DeviceExtensionSize = sizeof(device_extension_t);

    nt_ret = HidRegisterMinidriver(&hid_mdrv_regist);

    DBG_LEAVE(("%d", nt_ret));
    return nt_ret;
}


NTSTATUS DDKAPI device_add(IN PDRIVER_OBJECT drv_obj,
                           IN PDEVICE_OBJECT func_dev_obj)
{
    PDEVICE_OBJECT dev_obj;
    device_extension_t *dev_ext;

    DBG_ENTER(("%p, %p", drv_obj, func_dev_obj));

    /* device object */
    dev_obj = func_dev_obj;
    dev_obj->Flags &= ~DO_DEVICE_INITIALIZING;

    /* init devxce extension */
    dev_ext = GET_DEV_EXT(dev_obj);

    RtlZeroMemory(dev_ext, sizeof(device_extension_t));
    dev_ext->req_cnt = 1;
    KeInitializeEvent(&dev_ext->remove_event, SynchronizationEvent, FALSE);

    /* load settings */
    /* todo: implement */
    /*
    load_settings(dev_obj);
    */

    DBG_LEAVE(("STATUS_SUCCESS"));
    return STATUS_SUCCESS;
}


VOID DDKAPI driver_unload(IN PDRIVER_OBJECT drv_obj)
{
    DBG_ENTER(("%p", drv_obj));
    DBG_LEAVE(("void"));
    return;
}


NTSTATUS DDKAPI irp_create(IN PDEVICE_OBJECT dev_obj, IN PIRP irp)
{
    DBG_ENTER(("%p, %p", dev_obj, irp));

    COMPLETE_REQUEST(irp, 0, STATUS_SUCCESS, IO_NO_INCREMENT);

    DBG_LEAVE(("STATUS_SUCCESS"));
    return STATUS_SUCCESS;
}


NTSTATUS DDKAPI irp_close(IN PDEVICE_OBJECT dev_obj, IN PIRP irp)
{
    DBG_ENTER(("%p, %p", dev_obj, irp));

    COMPLETE_REQUEST(irp, 0, STATUS_SUCCESS, IO_NO_INCREMENT);

    DBG_LEAVE(("STATUS_SUCCESS"));
    return STATUS_SUCCESS;
}


NTSTATUS DDKAPI irp_pnp(IN PDEVICE_OBJECT dev_obj, IN PIRP irp)
{
    NTSTATUS nt_ret;
    device_extension_t *dev_ext;
    PIO_STACK_LOCATION irp_sp;
    UCHAR minor_func;

    DBG_ENTER(("%p, %p", dev_obj, irp));

    nt_ret = STATUS_SUCCESS;

    dev_ext = GET_DEV_EXT(dev_obj);
    irp_sp = IoGetCurrentIrpStackLocation(irp);
    minor_func = irp_sp->MinorFunction;

    nt_ret = inc_req_cnt(dev_ext);
    if(! NT_SUCCESS(nt_ret)) {
        COMPLETE_REQUEST(irp, 0, nt_ret, IO_NO_INCREMENT);
        DBG_LEAVE(("%d", nt_ret));
        return nt_ret;
    }

    {
        pnp_mn_func_t func;

        func = (pnp_mn_func_t)assq_pair(pnp_mn_func_list, minor_func, NULL);
        if(func != NULL) {
            nt_ret = func(dev_obj, irp);
        } else {
            IoSkipCurrentIrpStackLocation(irp);
            nt_ret = IoCallDriver(GET_NEXT_DEVICE_OBJECT(dev_obj), irp);
        }
    }

    if(minor_func != IRP_MN_REMOVE_DEVICE) {
        dec_req_cnt(dev_ext);
    }

    DBG_LEAVE(("%d", nt_ret));
    return nt_ret;
}


NTSTATUS DDKAPI irp_power(IN PDEVICE_OBJECT dev_obj, IN PIRP irp)
{
    NTSTATUS nt_ret;
    device_extension_t *dev_ext;

    DBG_ENTER(("%p, %p", dev_obj, irp));

    dev_ext = GET_DEV_EXT(dev_obj);

    nt_ret = inc_req_cnt(dev_ext);
    if(! NT_SUCCESS(nt_ret)) {
        COMPLETE_REQUEST(irp, 0, nt_ret, IO_NO_INCREMENT);
        DBG_LEAVE(("%d", nt_ret));
        return nt_ret;
    }

    IoSkipCurrentIrpStackLocation(irp);
    PoStartNextPowerIrp(irp);
    nt_ret = PoCallDriver(GET_NEXT_DEVICE_OBJECT(dev_obj), irp);

    dec_req_cnt(dev_ext);

    DBG_LEAVE(("%d", nt_ret));
    return nt_ret;
}


NTSTATUS DDKAPI irp_inter_ioctl(IN PDEVICE_OBJECT dev_obj, IN PIRP irp)
{
    NTSTATUS nt_ret;
    device_extension_t *dev_ext;

    DBG_ENTER(("%p, %p", dev_obj, irp));

    dev_ext = GET_DEV_EXT(dev_obj);

    irp->IoStatus.Information = 0;

    nt_ret = inc_req_cnt(dev_ext);
    if(NT_SUCCESS(nt_ret)) {
        PIO_STACK_LOCATION irp_sp;
        ULONG ioctl_code;
        ioctl_func_t func;

        irp_sp = IoGetCurrentIrpStackLocation(irp);
        ioctl_code = irp_sp->Parameters.DeviceIoControl.IoControlCode;

        func = (ioctl_func_t)
               assq_pair(ioctl_func_list, ioctl_code, NULL);

        if(func != NULL) {
            nt_ret = func(dev_obj, irp);
        } else {
            nt_ret = STATUS_NOT_SUPPORTED;
        }

        dec_req_cnt(dev_ext);
    }

    irp->IoStatus.Status = nt_ret;

    if(nt_ret == STATUS_PENDING) {
        IoMarkIrpPending(irp);
    } else {
        IoCompleteRequest(irp, IO_NO_INCREMENT);
        nt_ret = STATUS_SUCCESS;
    }

    DBG_LEAVE(("%d", nt_ret));
    return nt_ret;
}



NTSTATUS inc_req_cnt(IN device_extension_t *dev_ext)
{
    NTSTATUS nt_ret;
    LONG cnt;

    DBG_ENTER(("%p", dev_ext));

    cnt = InterlockedIncrement(&dev_ext->req_cnt);
    DBG_OUT(("increment ref: %d", cnt));

    if(dev_ext->removedp) {
        DBG_OUT(("already removed"));

        cnt = InterlockedDecrement(&dev_ext->req_cnt);
        DBG_OUT(("decrement ref: %d", cnt));
        if(cnt == 0) {
            KeSetEvent(&dev_ext->remove_event, IO_NO_INCREMENT, FALSE);
        }

        nt_ret = STATUS_DELETE_PENDING;
    } else {
        nt_ret = STATUS_SUCCESS;
    }

    DBG_LEAVE(("%d", nt_ret));
    return nt_ret;
}

VOID dec_req_cnt(IN device_extension_t *dev_ext)
{
    LONG cnt;

    DBG_ENTER(("%p", dev_ext));

    cnt = InterlockedDecrement(&dev_ext->req_cnt);
    DBG_OUT(("decrement ref: %d", cnt));
    if(cnt == 0) {
        KeSetEvent(&dev_ext->remove_event, IO_NO_INCREMENT, FALSE);
    }

    DBG_LEAVE(("void"));
}



#define REGISTRY_BASE_PATH L"\\Registry\\Machine\\Software\\noname\\hfsd"

#include <initguid.h>
#ifdef GUID_EXT
#undef GUID_EXT
#define GUID_EXT
#endif
DEFINE_GUID(guid_devinterface_hid,
            0x4d1e55b2l, 0xf16f, 0x11cf,
            0x88, 0xcb, 0x00, 0x11, 0x11, 0x00, 0x00, 0x30);

static NTSTATUS get_registry_dword(IN PWSTR subkey, IN PWSTR name,
                                   OUT PDWORD val, IN DWORD def_val)
{
    RTL_QUERY_REGISTRY_TABLE query[3];
    NTSTATUS nt_ret;

    DBG_ENTER(("\"%ws\", \"%ws\", %p, %u", subkey, name, val, def_val));

    RtlZeroMemory(query, sizeof(query));

    query[0].Flags = RTL_QUERY_REGISTRY_SUBKEY;
    query[0].Name = subkey;
    query[0].EntryContext = NULL;

    query[1].Flags = RTL_QUERY_REGISTRY_DIRECT;
    query[1].Name = name;
    query[1].EntryContext = val;
    query[1].DefaultType = REG_DWORD;
    query[1].DefaultData = &def_val;
    query[1].DefaultLength = sizeof(DWORD);

    nt_ret = RtlQueryRegistryValues(RTL_REGISTRY_ABSOLUTE,
                                    REGISTRY_BASE_PATH,
                                    query,
                                    NULL, NULL);

    DBG_LEAVE(("%d, %u", nt_ret, *val));
    return nt_ret;
}

static VOID load_settings_from_subkey(IN PDEVICE_OBJECT dev_obj,
                                      IN PWSTR subkey)
{
    device_extension_t *dev_ext;

    DBG_ENTER(("%p, \"%ws\"", dev_obj, subkey));

    dev_ext = GET_DEV_EXT(dev_obj);

    /* load DWORD setting */
    {
        struct {
            PWSTR name;
            PDWORD ptr;
            DWORD def_val;
        } *lptr, name_ptr_list[] = {

            {NULL, NULL, 0}
        };

        for(lptr = name_ptr_list; lptr->name != NULL; lptr++) {
            get_registry_dword(subkey, lptr->name, lptr->ptr, lptr->def_val);
        }
    }

    DBG_LEAVE(("void"));
}

static VOID load_settings_from_path(IN PDEVICE_OBJECT dev_obj, IN PWSTR path)
{
    DWORD val;
    UNICODE_STRING id_ubuf, subkey_ubuf;
    WCHAR id_buf[64], subkey_buf[256];

    DBG_ENTER(("%p, \"%ws\"", dev_obj, path));

    /* get device path -> ID mapping */
    get_registry_dword(L"map", path, &val, 0);
    if(val == 0) {
        DBG_LEAVE(("void: no mapping"));
        return;
    }

    /* convert ID to UNICODE_STRING */
    RtlZeroMemory(id_buf, sizeof(id_buf));
    id_ubuf.Length = 0;
    id_ubuf.MaximumLength = sizeof(id_buf);
    id_ubuf.Buffer = id_buf;

    RtlIntegerToUnicodeString(val, 16, &id_ubuf);

    /* prepend "parameter\" */
    RtlZeroMemory(subkey_buf, sizeof(subkey_buf));
    subkey_ubuf.Length = 0;
    subkey_ubuf.MaximumLength = sizeof(subkey_buf);
    subkey_ubuf.Buffer = subkey_buf;

    RtlAppendUnicodeToString(&subkey_ubuf, L"parameter\\");
    RtlAppendUnicodeStringToString(&subkey_ubuf, &id_ubuf);

    /* load parameters */
    load_settings_from_subkey(dev_obj, subkey_ubuf.Buffer);

    DBG_LEAVE(("void"));
}

VOID load_settings(IN PDEVICE_OBJECT dev_obj)
{
    device_extension_t *dev_ext;
    PWSTR sym, ptr;
    UNICODE_STRING symlnk, alias;
    NTSTATUS nt_ret;

    DBG_ENTER(("%p", dev_obj));

    dev_ext = GET_DEV_EXT(dev_obj);

    nt_ret = IoGetDeviceInterfaces(&guid_devinterface_hid,
                                   dev_obj, 0,
                                   &sym);

    for(ptr = sym; ptr[0] != 0; ) {
        /* load parameters */
        load_settings_from_path(dev_obj, ptr);

        /* get alias of device path */
        RtlInitUnicodeString(&symlnk, ptr);

        alias.Length = 0;
        alias.MaximumLength = 0;
        alias.Buffer = NULL;

        nt_ret = IoGetDeviceInterfaceAlias(&symlnk, &guid_devinterface_hid,
                                           &alias);
        if(NT_SUCCESS(nt_ret)) {
            /* load parameters (alias path) */
            DBG_OUT(("alias path exists: \"%ws\"", alias.Buffer));
            load_settings_from_path(dev_obj, alias.Buffer);
        }

        RtlFreeUnicodeString(&alias);

        /* search next path */
        while(ptr[0] != 0) { ptr++; }
        ptr++;
    }

    ExFreePool(sym);

    DBG_LEAVE(("void"));
}
