/*
 * main.h  -- difinition of main.c, pnp.c, ioctl.c
 *
 * $Id: main.h,v 1.1 2004/02/08 01:15:22 hos Exp $
 *
 */


#pragma pack(push, 8)

#include "ntddk.h"
#include "usbdi.h"
#include "usbd_lib.h"
#include "usbioctl.h"
#include "hidclass.h"
#include "hid_port.h"

#include "report_data.h"

#include "util.h"


typedef signed char SCHAR;

/* input: interrupt */
struct controller_raw_input_ir {
    UCHAR stick_x;              /* stick (left - right, 0x00 - 0xff) */
    UCHAR stick_y;              /* stick (top - bottom, 0x00 - 0xff)  */
    UCHAR rudder;               /* rudder (left - right, 0x00 - 0xff) */
    UCHAR throttle;             /* throttle (top - bottom, 0x00 - 0xff) */
    UCHAR hat_x;                /* hat (left - right, 0x00 - 0xff) */
    UCHAR hat_y;                /* hat (top - bottom, 0x00 - 0xff) */
    UCHAR button_a;             /* button A (press - press, 0x00 - 0xff) */
    UCHAR button_b;             /* button B (press - press, 0x00 - 0xff) */
};

/* input: vendor request 0x00 */
struct controller_raw_input_vr_00 {
    UCHAR fire_c : 1;           /* button fire-c */
    UCHAR button_d : 1;         /* button D */
    UCHAR hat : 1;              /* hat press */
    UCHAR button_st : 1;        /* button ST */

    UCHAR dpad1_top : 1;        /* d-pad 1 top */
    UCHAR dpad1_right : 1;      /* d-pad 1 right */
    UCHAR dpad1_bottom : 1;     /* d-pad 1 bottom */
    UCHAR dpad1_left : 1;       /* d-pad 1 left */

    UCHAR reserved4 : 4;

    UCHAR reserved3 : 1;
    UCHAR launch : 1;           /* button lauch */
    UCHAR trigger : 1;          /* trigger */
    UCHAR reserved2 : 1;
};

/* input: vendor request 0x01 */
struct controller_raw_input_vr_01 {
    UCHAR reserved2 : 4;

    UCHAR dpad3_right : 1;      /* d-pad 3 right */
    UCHAR dpad3_middle : 1;     /* d-pad 3 middle */
    UCHAR dpad3_left : 1;       /* d-pad 3 left */
    UCHAR reserved1 : 1;

    UCHAR mode_select : 2;      /* mode select (M1 - M2 - M3, 2 - 1 - 3) */
    UCHAR reserved3 : 1;
    UCHAR button_sw1 : 1;       /* button sw-1 */

    UCHAR dpad2_top : 1;        /* d-pad 2 top */
    UCHAR dpad2_right : 1;      /* d-pad 2 right */
    UCHAR dpad2_bottom : 1;     /* d-pad 2 bottom */
    UCHAR dpad2_left : 1;       /* d-pad 2 left */
};

/* input */
struct controller_raw_input {
    struct controller_raw_input_ir ir;
    struct controller_raw_input_vr_00 vr_00;
    struct controller_raw_input_vr_01 vr_01;
};

/* output: vendor request 0x0c */
struct controller_raw_output_vr_0c {
    UCHAR viblation;            /* viblation (stop - full, 0x00 - 0xff) */
};

/* output */
struct controller_raw_output {
    struct controller_raw_output_vr_0c vr_0c;
};

struct device_extension {
    /* flags */
    BOOLEAN removedp;
    BOOLEAN startedp;
    BOOLEAN surprise_removedp;

    /* request count */
    LONG req_cnt;

    /* remove event */
    KEVENT remove_event;

    /* usb datas */
    USHORT report_size;
    USHORT vendor_id;
    USHORT product_id;

    /* usb handles */
    USBD_CONFIGURATION_HANDLE usb_conf_handle;
    USBD_PIPE_HANDLE usb_input_pipe;
    USBD_PIPE_HANDLE usb_output_pipe;

    /* in/out state and its lock */
    struct controller_raw_input raw_input;
    struct controller_input_report input_state;
    struct controller_output_report output_state;
    KSPIN_LOCK state_lock;

    /* config settings */
};
typedef struct device_extension device_extension_t;


#define GET_HID_DEV_EXT(dev_obj) \
  ((PHID_DEVICE_EXTENSION)(dev_obj)->DeviceExtension)
#define GET_NEXT_DEVICE_OBJECT(dev_obj) \
  (GET_HID_DEV_EXT(dev_obj)->NextDeviceObject)
#define GET_DEV_EXT(dev_obj) \
  ((device_extension_t *)GET_HID_DEV_EXT(dev_obj)->MiniDeviceExtension)

#define COMPLETE_REQUEST(irp, info, status, pri) \
  (irp)->IoStatus.Information = (info); \
  (irp)->IoStatus.Status = (status); \
  IoCompleteRequest((irp), (pri));


#ifndef USBD_TRANSFER_DIRECTION_IN
#define USBD_TRANSFER_DIRECTION_IN 0x01
#endif

#ifndef USBD_SHORT_TRANSFER_OK
#define USBD_SHORT_TRANSFER_OK 0x02
#endif


extern struct uint_ptr_pair pnp_mn_func_list[];
extern struct uint_ptr_pair ioctl_func_list[];

typedef NTSTATUS (* pnp_mn_func_t)(IN PDEVICE_OBJECT, IN PIRP);
typedef NTSTATUS (* ioctl_func_t)(IN PDEVICE_OBJECT, IN PIRP);


NTSTATUS device_init(IN PDEVICE_OBJECT dev_obj, IN PIRP irp);
NTSTATUS device_remove(IN PDEVICE_OBJECT dev_obj);

NTSTATUS usb_get_interrupt_transfer(IN PDEVICE_OBJECT dev_obj,
                                    IN PVOID buf, ULONG *buf_size);
NTSTATUS usb_put_interrupt_transfer(IN PDEVICE_OBJECT dev_obj,
                                    IN PVOID buf, ULONG *buf_size);

NTSTATUS usb_get_vendor_request(IN PDEVICE_OBJECT dev_obj,
                                IN PVOID buf, ULONG *buf_size,
                                IN UCHAR request, IN USHORT index);
NTSTATUS usb_put_vendor_request(IN PDEVICE_OBJECT dev_obj,
                                IN PVOID buf, ULONG *buf_size,
                                IN UCHAR request, IN USHORT index);

NTSTATUS hid_init(IN PDEVICE_OBJECT dev_obj);
NTSTATUS hid_get_rep_desc(IN PDEVICE_OBJECT dev_obj, BYTE *buf, ULONG *size);
NTSTATUS hid_read_rep(IN PDEVICE_OBJECT dev_obj,
                      OUT BYTE *buf, OUT ULONG *size);
NTSTATUS hid_write_rep(IN PDEVICE_OBJECT dev_obj,
                       IN BYTE *buf, IN ULONG *size);
VOID hid_update_from_raw_input(IN PDEVICE_OBJECT dev_obj,
                               IN BYTE *buf, IN ULONG size);
VOID hid_update_feedback_light(PDEVICE_OBJECT dev_obj,
                               struct controller_raw_input *raw_in,
                               struct controller_input_report *in_state);
VOID hid_generate_raw_output(struct controller_output_report *out_state,
                             struct controller_raw_output *raw_out);


/* dbg: output debug info */
#define DBG_OUT(_x_)                            \
  KdPrint(("%s: %d: ", __FILE__, __LINE__));    \
  KdPrint(_x_);                                 \
  KdPrint(("\n"));

#define DBG_ENTER(_x_)                                                  \
  KdPrint(("%s: %d: enter: %s: ", __FILE__, __LINE__, __FUNCTION__));   \
  KdPrint(_x_);                                                         \
  KdPrint(("\n"));

#define DBG_LEAVE(_x_)                                                  \
  KdPrint(("%s: %d: leave: %s: ", __FILE__, __LINE__, __FUNCTION__));   \
  KdPrint(_x_);                                                         \
  KdPrint(("\n"));
